import 'package:flutter/material.dart';

/// Use extension method when available:
/// https://dart.dev/guides/language/extension-methods
/// Used by [MultiController] as a return type from [CreateController].
/// [Controller] is used for chaining .forward() of [AnimationController].
class Controller extends AnimationController {
  Controller({
    double value,
    duration,
    reverseDuration,
    lowerBound = 0.0,
    upperBound = 1.0,
    animationBehavior = AnimationBehavior.normal,
    @required TickerProvider vsync,
  }) : super(
          value: value,
          duration: duration,
          reverseDuration: reverseDuration,
          lowerBound: lowerBound,
          upperBound: upperBound,
          animationBehavior: animationBehavior,
          vsync: vsync,
        );

  /// Used to trigger another [Controller] when this one finishes.
  void chain(Controller otherController) {
    this.addStatusListener((status) {
      if (AnimationStatus.completed == status) {
        otherController.forward();
      }
    });
  }

  void chainList(List<Controller> otherControllers) {
    Controller previous = this;
    for (Controller current in otherControllers) {
      previous.chain(current);
      previous = current;
    }
  }
}

/// Remove when issue is fixed: https://github.com/dart-lang/sdk/issues/40588
/// Used with [CreateController] to pass parameters to create [Controller].
class ControllerData {
  final double value;
  final Duration duration;
  final Duration reverseDuration;
  final double lowerBound;
  final double upperBound;
  final AnimationBehavior animationBehavior;

  ControllerData({
    this.value,
    @required this.duration,
    this.reverseDuration,
    this.lowerBound = 0.0,
    this.upperBound = 1.0,
    this.animationBehavior,
  });
}

typedef Builder = Widget Function(
    BuildContext context, CreateController createController, Widget child);
typedef CreateController = Controller Function(
    ControllerData controllerCreator);

/// The main widget that is used within the widget tree to provide [Controller]s
class MultiController extends StatefulWidget {
  final Widget child;
  final Builder builder;

  MultiController({
    @required this.builder,
    this.child,
  });

  @override
  _MultiControllerState createState() => _MultiControllerState();
}

class _MultiControllerState extends State<MultiController>
    with TickerProviderStateMixin {
  List<AnimationController> controllers = [];

  @override
  void initState() {
    super.initState();
  }

  /// Creates a [Controller] and handles it's lifetime.
  Controller createController(ControllerData controllerData) {
    var controller = Controller(
      value: controllerData.value,
      duration: controllerData.duration,
      reverseDuration: controllerData.reverseDuration,
      lowerBound: controllerData.lowerBound,
      upperBound: controllerData.upperBound,
      animationBehavior: controllerData.animationBehavior,
      vsync: this,
    );
    controllers.add(controller);
    return controller;
  }

  @override
  Widget build(BuildContext context) {
    return widget.builder(
      context,
      (ControllerData controllerData) => createController(controllerData),
      widget.child,
    );
  }

  @override
  void dispose() {
    for (AnimationController controller in controllers) {
      controller.dispose();
    }
    super.dispose();
  }
}
